import importlib.util
from importlib.machinery import SourceFileLoader


def exec_module_from_filepath(filepath: str):
    # SourceFileLoader because spec_from_file_location only accepts files with a `py` file extension
    loader = SourceFileLoader(filepath, filepath)
    spec = importlib.util.spec_from_loader(filepath, loader)
    if spec is None:
        raise Exception(f"file @ {filepath} does not exist")
    
    module = importlib.util.module_from_spec(spec)
    loader.exec_module(module)

    return module


from plover.steno import Stroke
from plover.translation import Translator
import plover.log

from typing import Callable

from .Rule import Case, Condition, Clause, Statement, Prerequisite, LookupStrategy, Rule, StrokeFilter, Outline, Formatter, empty_stroke


class _ClauseBuilder:
    def __init__(
        self,
        stroke_filter: StrokeFilter,
        conditions: tuple[Condition, ...] = (),
    ):
        """A function that returns the strokes to test."""
        self.__stroke_filter = stroke_filter
        self.__conditions = conditions


    def __call__(self) -> Clause:
        return Clause(self.__stroke_filter, self.__conditions)
    

    def folds(self, *substrokes_steno: str):
        """Tests if ANY SINGLE given substroke is folded into a set of strokes."""

        substrokes = tuple(Stroke.from_steno(substroke_steno) for substroke_steno in substrokes_steno)
        cases: list[Case] = []
        for substroke in sorted(substrokes, key=lambda substroke: len(substroke), reverse=True):
            cases.append(Case(
                substroke,
                empty_stroke(),
            ))

        return _ClauseBuilder(
            self.__stroke_filter,
            self.__conditions + (Condition(tuple(cases)),),
        )

    fold = folds

    def toggles(self, *substrokes_steno: str):
        """Tests if ANY SINGLE given substroke is toggled in a set of strokes."""

        substrokes = tuple(Stroke.from_steno(substroke_steno) for substroke_steno in substrokes_steno)
        cases: list[Case] = []
        for substroke in sorted(substrokes, key=lambda substroke: len(substroke), reverse=True):
            cases.append(Case(
                empty_stroke(),
                substroke,
            ))

        return _ClauseBuilder(
            self.__stroke_filter,
            self.__conditions + (Condition(tuple(cases)),),
        )
    
    toggle = toggles

class _LookupStrategyGatherer:
    def __init__(self, prerequisite: Prerequisite):
        self.__prerequisite = prerequisite

    def then(self, *lookup_strategy_creators: Callable[[], LookupStrategy]):
        """Gathers a list of lookup strategies that are run IN ORDER and INDEPENDENTLY."""
        return Rule(self.__prerequisite, tuple(create_lookup_strategy() for create_lookup_strategy in lookup_strategy_creators))
    
    def and_also(self, *rules: Rule):
        return Rule(self.__prerequisite, ()).unless_also(*rules)
    

class _TranslationModificationGatherer:
    def __init__(
        self,
        modify_translation: Callable[[str], str]=lambda translation: translation,
        modify_outline: Callable[[Outline], Outline]=lambda strokes: strokes,
        check_additional_folds=True,
    ):
        self.__modify_outline = modify_outline
        self.__modify_translation = modify_translation
        self.__check_additional_folds = check_additional_folds


    def __call__(self) -> LookupStrategy:
        @LookupStrategy.of
        def handler(defolded_strokes: Outline, folds: Outline, strokes: Outline, translator: Translator):
            original_check_additional_folds = Rule.check_additional_folds
            if not self.__check_additional_folds:
                Rule.check_additional_folds = False

            foldless_translation = translator.lookup(self.__modify_outline(defolded_strokes))

            if not self.__check_additional_folds:
                Rule.check_additional_folds = original_check_additional_folds

            if foldless_translation is None:
                return None
                        
            return self.__modify_translation(foldless_translation)

        return handler
    

    def modify_translation(self, modify_translation: Formatter) -> "_TranslationModificationGatherer":
        """Translates the outline without any folds, and modifies the translation according to the callback."""
        return _TranslationModificationGatherer(lambda translation: modify_translation(self.__modify_translation(translation)), self.__modify_outline)
    
    def prefix_translation(self, string: str) -> "_TranslationModificationGatherer":
        return self.modify_translation(lambda translation: f"{string}{translation}")
        
    def suffix_translation(self, string: str) -> "_TranslationModificationGatherer":
        return self.modify_translation(lambda translation: f"{translation}{string}")


class _OutlineModificationGatherer:
    def __init__(self, modify_outline: Callable[[Outline], Outline]=lambda strokes: strokes, check_additional_folds=True):
        self.__modify_outline = modify_outline
        self.__check_additional_folds = check_additional_folds


    def __call__(self) -> LookupStrategy:
        return self.__to_translation_modification_gatherer()()


    def __to_translation_modification_gatherer(self, modify_translation: Formatter=lambda translation: translation) -> _TranslationModificationGatherer:
        return _TranslationModificationGatherer(modify_translation, self.__modify_outline, self.__check_additional_folds)

    def modify_translation(self, modify_translation: Formatter=lambda translation: translation) -> _TranslationModificationGatherer:
        """Translates the outline without any folds, and modifies the translation according to the callback."""
        return self.__to_translation_modification_gatherer(modify_translation)
    
    def prefix_translation(self, string: str) -> _TranslationModificationGatherer:
        return self.__to_translation_modification_gatherer().prefix_translation(string)
        
    def suffix_translation(self, string: str) -> _TranslationModificationGatherer:
        return self.__to_translation_modification_gatherer().suffix_translation(string)
    
    def check_additional_folds(self) -> _TranslationModificationGatherer:
        return _OutlineModificationGatherer(self.__modify_outline, True).__to_translation_modification_gatherer()
    
    def modify_outline(self, modify_outline: Callable[[Outline], Outline]) -> "_OutlineModificationGatherer":
        """Translates the outline without any folds, and modifies the translation according to the callback."""
        return _OutlineModificationGatherer(lambda strokes: modify_outline(self.__modify_outline(strokes)), False)
    
    def prefix_outline(self, new_strokes_steno: str) -> "_OutlineModificationGatherer":
        new_strokes = tuple(Stroke.from_steno(stroke_steno) for stroke_steno in new_strokes_steno.split("/"))
        return self.modify_outline(lambda strokes: new_strokes[:-1] + (new_strokes[-1] + strokes[0],) + strokes[1:])
        
    def suffix_outline(self, new_strokes_steno: str) -> "_OutlineModificationGatherer":
        new_strokes = tuple(Stroke.from_steno(stroke_steno) for stroke_steno in new_strokes_steno.split("/"))
        return self.modify_outline(lambda strokes: strokes[:-1] + (new_strokes[0] + strokes[-1],) + new_strokes[1:])


def when(*builders: "Callable[[], Statement | Clause]") -> _LookupStrategyGatherer:
    """Gathers statements at least one of which must be true for this rule to be used."""
    return _LookupStrategyGatherer(Prerequisite(tuple(create() for create in builders)))
    
def when_all(*builders: "Callable[[], Statement | Clause]") -> _LookupStrategyGatherer:
    """Gathers statements all of which must be true for this rule to be used."""
    return when(all_rules(*builders))

first_stroke: _ClauseBuilder = _ClauseBuilder(lambda strokes: (strokes[0],))
last_stroke: _ClauseBuilder = _ClauseBuilder(lambda strokes: (strokes[-1],))
all_strokes: _ClauseBuilder = _ClauseBuilder(lambda strokes: strokes)
def nth_stroke(index: int):
    return _ClauseBuilder(lambda strokes: (strokes[index],))

def filtered_strokes(stroke_filter: StrokeFilter):
    return _ClauseBuilder(stroke_filter)


def all_rules(*builders: "Callable[[], Statement | Clause]"):
    return lambda: Statement(True, tuple(create() for create in builders))

def any_single_rule(*builders: "Callable[[], Statement | Clause]"):
    return lambda: Statement(False, tuple(create() for create in builders))


def modify_translation(modify_translation: Formatter=lambda translation: translation) -> _TranslationModificationGatherer:
    """Translates the outline without any folds, and modifies the translation according to the callback."""
    return _TranslationModificationGatherer(modify_translation)

def prefix_translation(string: str) -> _TranslationModificationGatherer:
    return _TranslationModificationGatherer().prefix_translation(string)

def suffix_translation(string: str) -> _TranslationModificationGatherer:
    return _TranslationModificationGatherer().suffix_translation(string)

def modify_outline(modify_outline: Callable[[Outline], Outline]) -> _OutlineModificationGatherer:
    """Translates the outline without any folds, and modifies the outline according to the callback."""
    return _OutlineModificationGatherer(modify_outline)

def prefix_outline(new_strokes_steno: str) -> _OutlineModificationGatherer:
    return _OutlineModificationGatherer().prefix_outline(new_strokes_steno)

def suffix_outline(new_strokes_steno: str) -> _OutlineModificationGatherer:
    return _OutlineModificationGatherer().suffix_outline(new_strokes_steno)

def group_rules(*rules: Rule) -> Rule:
    return Rule(Prerequisite(()), (), rules)


def unfold_suffix():
    """Default folding behavior. Removes the fold from the final stroke and appends the folded chord as a new
    stroke, using the chord's dictionary entry."""

    @LookupStrategy.of
    def handler(defolded_strokes: Outline, folds: Outline, strokes: Outline, translator: Translator):
        foldless_translation = translator.lookup(defolded_strokes)
        if foldless_translation is None:
            return None
        
        fold_chord_translation = translator.lookup((folds[-1],))
        if fold_chord_translation is None:
            return None
        
        return f"{foldless_translation} {fold_chord_translation}"

    return handler

def unfold_prefix():
    @LookupStrategy.of
    def handler(defolded_strokes: Outline, folds: Outline, strokes: Outline, translator: Translator):
        foldless_translation = translator.lookup(defolded_strokes)
        if foldless_translation is None:
            return None
        
        fold_chord_translation = translator.lookup((folds[0],))
        if fold_chord_translation is None:
            return None
    
        return f"{fold_chord_translation} {foldless_translation}"
    
    return handler


def use_defolded_translation():
    @LookupStrategy.of
    def handler(defolded_strokes: Outline, folds: Outline, strokes: Outline, translator: Translator):
        return translator.lookup(defolded_strokes)
    
    return handler

def default_system_rules():
    """Plover's default folding rules, depend on the current system."""

    from plover.system import SUFFIX_KEYS # type: ignore

    return when(last_stroke.folds(*SUFFIX_KEYS)).then(unfold_suffix)


from plover.engine import StenoEngine
from plover.translation import Translator
from plover.steno import Stroke

from .lib.builder import Rule

class TranslatorContainer:
    checking_shorter_outlines = False
    __shorter_outline_length = -1
    __checked_shorter_outlines = False


    def __init__(self):
        self.translator: "Translator | None" = None

    def check_shorter_outlines(self, longest_key_len: int):
        if self.translator is None: raise Exception("EngineGetterExtension is not enabled; enable it in Plover config > `Plugins`")

        # Check shorter outlines before trying to defold (mimic default Plover behavior)
        if TranslatorContainer.checking_shorter_outlines: return -1

        if TranslatorContainer.__checked_shorter_outlines:
            return self.__shorter_outline_length

        TranslatorContainer.__checked_shorter_outlines = True

        TranslatorContainer.checking_shorter_outlines = True
        strokes = []
        for translation in self.translator.get_state().translations:
            for stroke in translation.strokes:
                strokes.append(stroke)
                
                if len(strokes) == longest_key_len:
                    break
            if len(strokes) == longest_key_len:
                break

        for start_index in range(len(strokes) - 1, 0, -1):
            shorter_translation = self.translator.lookup(strokes[start_index:])
            if shorter_translation is None: continue

            TranslatorContainer.checking_shorter_outlines = False
            self.__shorter_outline_length = True
            return len(strokes) - start_index
        TranslatorContainer.checking_shorter_outlines = False

        self.__shorter_outline_length = False
        return -1
    
    def setup(self, engine: StenoEngine):
        self.translator = engine._translator

        def on_stroked(stroke: Stroke):
            Rule.clear_unmatched_rules()
            TranslatorContainer.clear_shorter_outlines()
        engine.hook_connect("stroked", on_stroked)

    @staticmethod
    def clear_shorter_outlines():
        TranslatorContainer.__shorter_outline_length = -1
        TranslatorContainer.__checked_shorter_outlines = False



translator_container = TranslatorContainer()

class EngineGetterExtension:
    def __init__(self, engine: StenoEngine):
        translator_container.setup(engine)
    
    def start(self):
        pass

    def stop(self):
        pass


from typing import Optional

from plover.steno import Stroke
from plover.steno_dictionary import StenoDictionary

from .lib.util import exec_module_from_filepath


class PythonDictionary(StenoDictionary):
    readonly = True


    def __init__(self):
        super().__init__()

        """(override)"""
        self._longest_key = 0
        self.__lookup = None
        self.__reverse_lookup = lambda translation: set()


    def _load(self, filepath: str):
        module = exec_module_from_filepath(filepath)

        self.__lookup = module.lookup
        self._longest_key = module.LONGEST_KEY
        self.__reverse_lookup = getattr(module, "reverse_lookup", self.__reverse_lookup)

    def __getitem__(self, key: tuple[str, ...]) -> str:
        result = self.__lookup_steno(key)
        if result is None:
            raise KeyError
        
        return result

    def get(self, key: tuple[str, ...], fallback=None) -> Optional[str]:
        result = self.__lookup_steno(key)
        if result is None:
            return fallback
        
        return result
    
    def __lookup_steno(self, key: tuple[str, ...]) -> Optional[str]:
        if self.__lookup is None:
            raise Exception("tried looking up before dictionary was loaded")
    
        if len(key) > self._longest_key:
            return None

        strokes = tuple(Stroke.from_steno(steno) for steno in key)
        return self.__lookup(strokes)
    
    def reverse_lookup(self, translation: str) -> set[str]:
        return set(self.__reverse_lookup(translation))


from typing import Optional

from plover.steno import Stroke
from plover.steno_dictionary import StenoDictionary

from .EngineGetterExtension import translator_container, TranslatorContainer
from .lib.builder import Rule
from .lib.util import exec_module_from_filepath


class PythonFoldingDictionary(StenoDictionary):
    readonly = True

    def __init__(self):
        super().__init__()

        """(override)"""
        self._longest_key = 8
        self.__rules: "list[Rule] | None" = None
        self.__current_rules: set[Rule] = set()

    def _load(self, filepath: str):
        module = exec_module_from_filepath(filepath)

        self.__rules = module.rules
        self._longest_key = getattr(module, "LONGEST_KEY", self._longest_key)

    def __getitem__(self, key: tuple[str, ...]) -> str:
        result = self.__lookup(key)
        if result is None:
            raise KeyError
        
        return result

    def get(self, key: tuple[str, ...], fallback=None) -> Optional[str]:
        result = self.__lookup(key)
        if result is None:
            return fallback
        
        return result
    
    def __lookup(self, key: tuple[str, ...]) -> Optional[str]:
        if self.__rules is None:
            raise Exception("tried looking up before dictionary was loaded")
        if translator_container.translator is None:
            raise Exception("EngineGetterExtension is not enabled; enable it in Plover config > `Plugins`")
        
        if not Rule.check_additional_folds:
            return None
        
        if TranslatorContainer.checking_shorter_outlines:
            return None
        
        strokes = tuple(Stroke.from_steno(steno) for steno in key)

        shorter_outline_length = translator_container.check_shorter_outlines(self._longest_key)
        shorter_outline_found = shorter_outline_length != -1 and shorter_outline_length < len(strokes)


        for rule in self.__rules:
            # Prevent a rule from looking up itself
            if rule in self.__current_rules:
                continue
            
            self.__current_rules.add(rule)

            translation = rule(strokes, translator_container.translator, shorter_outline_found)
            
            self.__current_rules.remove(rule)
            
            if translation is None:
                continue
            return translation

        return None


from plover.steno import Stroke
from plover.translation import Translator
import plover.log

from typing import Callable, Generator, TypeVar


T = TypeVar("T")

Lookup = Callable[[tuple[Stroke], Translator], str]
Formatter = Callable[[str], str]

Outline = tuple[Stroke, ...]
StrokeFilter = Callable[[Outline], Outline]

def _toggle_substroke(stroke: Stroke, substroke: Stroke):
    if substroke in stroke:
        return stroke - substroke
    else:
        return stroke + substroke
    
def _strokes_overlap(a: Stroke, b: Stroke):
    return ~(~a | ~b) != 0  # De Morgan's Law

def empty_stroke() -> Stroke:
    return Stroke.from_keys(())

def _create_stroke_index_mapping(strokes: Outline):
    return {
        stroke: i
        for i, stroke in enumerate(strokes)
    }


def _all_combinations(options: "tuple[tuple[T, ...], ...]", *, last_index=-1) -> "Generator[tuple[T, ...], None, None]":
    if len(options) + last_index < 0:
        yield ()
        return

    for option in options[last_index]:
        for combination in _all_combinations(options, last_index=last_index - 1):
            yield (*combination, option)

def _case_groups(options: "tuple[tuple[Case, ...], ...]", stroke_filter: StrokeFilter):
    yield from (_CaseGroup(*((case, stroke_filter) for case in cases)) for cases in _all_combinations(options))

class Case:
    def __init__(
        self,
        contained_substroke: Stroke,
        toggled_substroke: Stroke,
    ):
        self.__contained_substroke = contained_substroke
        self.__toggled_substroke = toggled_substroke

    def satisfied_by(self, filtered_strokes: Outline):
        return all(self.__contained_substroke in stroke for stroke in filtered_strokes)
    
    def remove_fold(self, stroke: Stroke):
        return _toggle_substroke(stroke - self.__contained_substroke, self.__toggled_substroke)
    
    @property
    def fold_substroke(self):
        return self.__contained_substroke + self.__toggled_substroke
    
    def __repr__(self):
        return f"_Case(contained={self.__contained_substroke}, toggled={self.__toggled_substroke})"

class Condition:
    def __init__(
        self,
        cases: tuple[Case, ...]
    ):
        self.__cases = cases
    
    def satisifed_cases(self, filtered_strokes: Outline):
        """Returns the case which is satisfied by the given outline, or None if no such case exists."""
        for case in self.__cases:
            if case.satisfied_by(filtered_strokes):
                yield case
    
    def __repr__(self):
        return f"_Condition(cases={self.__cases})"

class _CaseGroup:
    def __init__(self, *cases_and_filters: tuple[Case, StrokeFilter]):
        self.__cases_and_filters = cases_and_filters

    def __iter__(self):
        return self.__cases_and_filters.__iter__()

    def remove_folds(self, strokes: Outline, *, stroke_index_mapping: "dict[Stroke, int] | None"=None):
        new_strokes = list(strokes)
        stroke_index_mapping = stroke_index_mapping or _create_stroke_index_mapping(strokes)
        for case, stroke_filter in self.__cases_and_filters:
            for stroke in stroke_filter(strokes):
                new_strokes[stroke_index_mapping[stroke]] = case.remove_fold(new_strokes[stroke_index_mapping[stroke]])

        return tuple(new_strokes)
    
    def get_folds(self, strokes: Outline, *, folds: "Outline | None"=None, stroke_index_mapping: "dict[Stroke, int] | None"=None):
        if folds is not None:
            new_folds = list(folds)
        else:
            new_folds = [empty_stroke() for _ in strokes]
        
        stroke_index_mapping = stroke_index_mapping or _create_stroke_index_mapping(strokes)
        for case, stroke_filter in self.__cases_and_filters:
            for stroke in stroke_filter(strokes):
                new_folds[stroke_index_mapping[stroke]] += case.fold_substroke

        return tuple(new_folds)
    
    def merge(self, case_group: "_CaseGroup"):
        return _CaseGroup(*self.__cases_and_filters, *case_group.__cases_and_filters)
    
    def __repr__(self):
        return f"_CaseGroup(cases={self.__cases_and_filters})"
    
class Clause:
    """Determines whether a substroke has been folded into a stroke, and what keys comprise that fold."""

    def __init__(
        self,
        stroke_filter: StrokeFilter,
        conditions: tuple[Condition, ...],
    ):
        """A function that returns the strokes to test."""
        self.__stroke_filter = stroke_filter
        self.__conditions = conditions

        # assert not _strokes_overlap(contained_substroke, toggled_substroke)

    def satisfied_case_groups(self, strokes: Outline):
        """Generates all groups of cases which are satisfied by the given outline for each condition."""

        case_combinations = tuple(tuple(condition.satisifed_cases(self.__stroke_filter(strokes))) for condition in self.__conditions)
        for case_group in _case_groups(case_combinations, self.__stroke_filter):
            yield case_group
        
    def __repr__(self):
        return f"_Clause(conditions={self.__conditions})"


class Statement:
    def __init__(self, conjunctive: bool, substatements: "tuple[Statement | Clause, ...]"):
        self.__conjunctive = conjunctive
        self.__substatements = substatements

    def satisfied_case_groups(self, strokes: Outline):
        """Generates all groups of cases which are satisfied by the given outline for each condition."""

        if self.__conjunctive:
            case_group_sequences = tuple(tuple(substatement.satisfied_case_groups(strokes)) for substatement in self.__substatements)

            running_case_group = _CaseGroup()
            for case_group_sequence in _all_combinations(case_group_sequences):
                for case_group in case_group_sequence:
                    running_case_group = running_case_group.merge(case_group)
            
            yield running_case_group
            return

        for substatement in self.__substatements:
            satisfied = False
            
            for case_group in substatement.satisfied_case_groups(strokes):
                satisfied = True
                yield case_group

            if satisfied:
                return

class Prerequisite:
    """Bundles a group of `_Statement`s."""

    def __init__(self, elements: "tuple[Statement | Clause, ...]"):
        self.__elements = elements
    
    def satisfied_folds(self, strokes: Outline):
        """Generates all groups of cases which are satisfied by the given outline for each clause."""
        stroke_index_mapping = _create_stroke_index_mapping(strokes)
        
        for element in self.__elements:
            for case_group in element.satisfied_case_groups(strokes):
                defolded_strokes = case_group.remove_folds(strokes, stroke_index_mapping=stroke_index_mapping)
                folds = case_group.get_folds(strokes, stroke_index_mapping=stroke_index_mapping)

                yield defolded_strokes, folds


_LookupStrategyHandler = Callable[[Outline, Outline, Outline, Translator], "str | None"]

class LookupStrategy:
    """Manipulates the input strokes and folded chords, queries the main translator, and produces new translations."""

    def __init__(self, handler: _LookupStrategyHandler):
        self.__handler = handler

    @staticmethod
    def of(fn: _LookupStrategyHandler):
        return LookupStrategy(fn)
    
    def __call__(self, defolded_strokes: Outline, folds: Outline, strokes: Outline, translator: Translator) -> "str | None":
        return self.__handler(defolded_strokes, folds, strokes, translator)


class Rule:
    def __init__(
        self,
        prerequisite: Prerequisite,
        lookup_strategies: tuple[LookupStrategy, ...],
        additional_rules: "tuple[Rule, ...]"=(),
        alternative_rules: "tuple[Rule, ...]"=(),
        allow_shorter_outline: bool=False,
    ):
        self.__prerequisite = prerequisite
        self.__lookup_strategies = lookup_strategies
        self.__additional_rules = additional_rules
        self.__alternative_rules = alternative_rules
        self.__allow_shorter_outline = allow_shorter_outline


    check_additional_folds = True
    shorter_outline_found = False

    __current_folds: "tuple[Stroke, ...] | None" = None
    __unmatched_rules: "dict[Outline, set[Rule]]" = {}  # for memoization
    @classmethod
    def clear_unmatched_rules(cls):
        cls.__unmatched_rules.clear()

    def __call__(self, strokes: Outline, translator: Translator, shorter_outline_found: bool) -> "str | None":
        if strokes in Rule.__unmatched_rules and self in Rule.__unmatched_rules[strokes]:
            return None

        if len(strokes[0]) == 0:
            return None
        
        if shorter_outline_found and not self.__allow_shorter_outline:
            return None

        for defolded_strokes, folds in self.__prerequisite.satisfied_folds(strokes):
            # Ensure that folds between rules do not overlap
            last_folds = Rule.__current_folds
            if Rule.__current_folds is None:
                Rule.__current_folds = folds
            else:
                new_folds: list[Stroke] = []
                overlap_found = False
                for i, fold in enumerate(Rule.__current_folds):
                    if _strokes_overlap(fold, folds[i]):
                        overlap_found = True
                        break
                    
                    new_folds.append(fold + folds[i])

                if overlap_found:
                    continue

                Rule.__current_folds = tuple(new_folds)

            # Removing folds will not affect rules that already don't match
            if strokes in Rule.__unmatched_rules:
                Rule.__unmatched_rules[defolded_strokes] = set(Rule.__unmatched_rules[strokes])
            else:
                Rule.__unmatched_rules[defolded_strokes] = set()

            # Check additional rules before the main rule
            for additional_rule in self.__additional_rules:
                translation = additional_rule(defolded_strokes, translator, shorter_outline_found)
                if translation is not None:
                    Rule.__current_folds = last_folds
                    return translation

            for lookup in self.__lookup_strategies:
                translation = lookup(defolded_strokes, folds, strokes, translator)
                if translation is not None:
                    Rule.__current_folds = last_folds
                    return translation

            for alternative_rule in self.__alternative_rules:
                translation = alternative_rule(defolded_strokes, translator, shorter_outline_found)
                if translation is not None:
                    Rule.__current_folds = last_folds
                    return translation
                
        
            Rule.__current_folds = last_folds
            del Rule.__unmatched_rules[defolded_strokes]

        if strokes in Rule.__unmatched_rules:
            Rule.__unmatched_rules[strokes].add(self)
        else:
            Rule.__unmatched_rules[strokes] = {self}

        return None
    
    def or_also(self, *alternative_rules: "Rule"):
        return Rule(self.__prerequisite, self.__lookup_strategies, self.__additional_rules, self.__alternative_rules + alternative_rules, self.__allow_shorter_outline)
    
    def unless_also(self, *additional_rules: "Rule"):
        return Rule(self.__prerequisite, self.__lookup_strategies, self.__additional_rules + additional_rules, self.__alternative_rules, self.__allow_shorter_outline)
    
    def preferring_folds(self):
        return Rule(self.__prerequisite, self.__lookup_strategies, self.__additional_rules, self.__alternative_rules, True)
